@@ -1,5 +1,7 @@
+from urllib3 import ProxyManager
 from urllib3 import PoolManager
 from urllib3 import Timeout
+from functools import wraps
 import json
 import os
 import tarfile
@@ -8,6 +10,7 @@ from .rest import simple_delete
 from .rest import simple_post
 from .rest import multipart_post
 import numpy as np
+# urllib3.disable_warnings(urllib3.exceptions.InsecureRequestWarning)
 
 __all__ = ['BoonException', 'NanoHandle']
 
@@ -34,6 +37,23 @@ class NanoHandle:
             license_file (str): path to .BoonLogic license file
             timeout (float): read timeout for http requests
 
+        Environment:
+            BOON_LICENSE_FILE: sets license_file path
+            BOON_LICENSE_ID: sets license_id
+            BOON_API_KEY: overrides the api-key as found in .BoonLogic.license file
+            BOON_API_TENANT: overrides the api-tenant as found in .BoonLogic.license file
+            BOON_SERVER: overrides the server as found in .BoonLogic.license file
+            PROXY_SERVER: overrides the proxy server as found in .BoonLogic.license file
+
+        Example:
+            ```python
+            try:
+                nano = bn.NanoHandle()
+            except bn.BoonException as be:
+                print(be)
+                sys.exit(1)
+            ```
+
         """
         self.license_file = license_file
         self.license_id = None
@@ -104,6 +124,12 @@ class NanoHandle:
                     "'api-tenant' is missing from configuration, set via BOON_TENANT or in ~/.BoonLogic.license file")
             self.api_tenant = license_block['api-tenant']
 
+        # load the https proxy (if any)
+        self.proxy_server = os.getenv('PROXY_SERVER')
+        if not self.proxy_server:
+            if 'proxy-server' in license_block.keys():
+                self.proxy_server = license_block['proxy-server']
+
         # set up base url
         self.url = self.server + '/expert/v3/'
         if "http" not in self.server:
@@ -111,7 +137,20 @@ class NanoHandle:
 
         # create pool manager
         timeout_inst = Timeout(connect=30.0, read=timeout)
-        self.http = PoolManager(timeout=timeout_inst)
+        if self.proxy_server:
+            # proxy pool
+            self.http = ProxyManager(self.proxy_server, maxsize=10, timeout=timeout_inst)
+        else:
+            # non-proxy pool
+            self.http = PoolManager(timeout=timeout_inst)
+
+    def _is_configured(f):
+        @wraps(f)
+        def inner(*args, **kwargs):
+            if args[0].numeric_format not in ['int16', 'uint16', 'float32']:
+                return False, "nano instance is not configured"
+            return f(*args, **kwargs)
+        return inner
 
     def open_nano(self, instance_id):
         """Creates or attaches to a nano pod instance
@@ -152,9 +191,12 @@ class NanoHandle:
         self.http.clear()
         return result, None
 
-    def create_config(self, feature_count, numeric_format, min_val=np.array([0]), max_val=np.array([1]), weight=np.array([1]),
-                      percent_variation=.05, streaming_window=1, accuracy=.99, label=None):
-        """generate a configuration template for the given parameters
+    def create_config(self, feature_count, numeric_format, cluster_mode='batch', min_val=0, max_val=1,
+                      weight=1, label=None,
+                      percent_variation=.05, streaming_window=1, accuracy=.99,
+                      autotunePV=True, autotuneRange=True, autotune_by_feature=True, autotune_max_clusters=1000, exclusions=None,
+                      streaming_autotune=True, streaming_buffer=10000, learning_numerator=10, learning_denominator=10000, learning_max_clusters=1000, learning_samples=1000000):
+        """Generate a configuration template for the given parameters
 
         A discrete configuration is specified as a list of min, max, weights, and labels
 
@@ -168,60 +210,120 @@ class NanoHandle:
             max_val (list): corresponding maximum value, a single element list assigns all features with same max_val
             weight (list): weight for this feature, a single element list assigns all features with same weight
             label (list): list of labels to assign to features
-            percent_variation (float):
-            streaming_window (integer):
-            accuracy (float):
+            percent_variation (float): amount of variation allowed within clusters
+            streaming_window (integer): number of consecutive vectors treated as one inference (parametric parameter)
+            accuracy (float): statistical accuracy of the clusters
+            autotunePV (bool): whether to autotune the percent variation
+            autotuneRange (bool): whether to autotune the min and max values
+            autotune_by_feature (bool): whether to have individually set min and max values for each feature
+            autotune_max_clusters (int): max number of clusters allowed
+            exclusions (list): features to exclude while autotuning
+            streaming_autotune (bool): whether to autotune while in streaming mode
+            streaming_buffer (int): number of samples to autotune on
+            learning_numerator (int): max number of new clusters learned
+            learning_denominator (int): number of samples over which the new clusters are learned
+            learning_max_clusters (int): max number of clusters before turning off learning
+            learning_samples (int): max number of samples before turning off learning
+
 
         Returns:
             result (boolean): true if successful (configuration was successfully created)
             response (dict or str): configuration dictionary when result is true, error string when result is false
 
         """
-        template_cmd = self.url + 'configTemplate/' + '?api-tenant=' + self.api_tenant
-        template_cmd += '&featureCount=' + str(feature_count)
-        template_cmd += '&numericFormat=' + str(numeric_format)
-        if isinstance(min_val, list):
-            template_cmd += '&minVal=' + ",".join([str(s) for s in min_val])
-        elif isinstance(min_val, np.ndarray):
-            template_cmd += '&minVal=' + ",".join([str(s) for s in min_val])
+
+        if isinstance(min_val, int) or isinstance(min_val, float):
+            min_val = np.array([min_val] * feature_count)
+        if isinstance(max_val, int) or isinstance(max_val, float):
+            max_val = np.array([max_val] * feature_count)
+        if isinstance(weight, int):
+            weight = np.array([weight] * feature_count)
+
+        config = {}
+        config['clusterMode'] = cluster_mode
+        config['numericFormat'] = numeric_format
+        config['features'] = []
+
+        if (isinstance(min_val, list) or isinstance(min_val, np.ndarray)) and (isinstance(max_val, list) or isinstance(max_val, np.ndarray)) and (isinstance(weight, list) or isinstance(weight, np.ndarray)):
+            if len(min_val) != len(max_val) or len(min_val) != len(weight):
+                return False, "parameters must be lists of the same length"
+
+            for min, max, w in zip(min_val, max_val, weight):
+                tempDict = {}
+                if numeric_format == 'float32':
+                    tempDict['minVal'] = float(min)
+                    tempDict['maxVal'] = float(max)
+                else:
+                    tempDict['minVal'] = int(min)
+                    tempDict['maxVal'] = int(max)
+                tempDict['weight'] = int(w)
+                config['features'].append(tempDict)
         else:
-            return False, "min_val must be list or numpy array"
-        if isinstance(max_val, list):
-            template_cmd += '&maxVal=' + ",".join([str(s) for s in max_val])
-        elif isinstance(max_val, np.ndarray):
-            template_cmd += '&maxVal=' + ",".join([str(s) for s in max_val])
-        else:
-            return False, "max_val must be list or numpy array"
-        if isinstance(weight, list):
-            template_cmd += '&weight=' + ",".join([str(s) for s in weight])
-        elif isinstance(weight, np.ndarray):
-            template_cmd += '&weight=' + ",".join([str(s) for s in weight])
-        else:
-            return False, "weight must be list or numpy array"
+            return False, "min_val, max_val and weight must be list or numpy array"
+
         if isinstance(label, list):
-            template_cmd += '&label=' + ",".join([str(s) for s in label])
+            if len(label) != len(min_val):
+                return False, "label must be the same length as other parameters"
+            for i, l in enumerate(label):
+                config['features'][i]['label'] = l
         elif label:
             return False, "label must be list"
-        template_cmd += '&percentVariation=' + str(percent_variation)
-        template_cmd += '&streamingWindowSize=' + str(streaming_window)
-        template_cmd += '&accuracy=' + str(accuracy)
 
-        return simple_get(self, template_cmd)
+        config['percentVariation'] = percent_variation
+        config['streamingWindowSize'] = streaming_window
+        config['accuracy'] = accuracy
 
-    def configure_nano(self, config):
-        """returns the posted clustering configuration
+        config['autoTuning'] = {}
+        config['autoTuning']['autoTunePV'] = autotunePV
+        config['autoTuning']['autoTuneRange'] = autotuneRange
+        config['autoTuning']['autoTuneByFeature'] = autotune_by_feature
+        config['autoTuning']['maxClusters'] = autotune_max_clusters
+        if isinstance(exclusions, list):
+            config['autoTuning']['exclusions'] = exclusions
+        elif exclusions:
+            return False, 'exclusions must be a list'
+
+        config['streaming'] = {}
+        config['streaming']['enableAutoTuning'] = streaming_autotune
+        config['streaming']['samplesToBuffer'] = streaming_buffer
+        config['streaming']['learningRateNumerator'] = learning_numerator
+        config['streaming']['learningRateDenominator'] = learning_denominator
+        config['streaming']['learningMaxClusters'] = learning_max_clusters
+        config['streaming']['learningMaxSamples'] = learning_samples
+
+        return config
+
+    def configure_nano(self, feature_count, numeric_format, cluster_mode='batch', min_val=0, max_val=1,
+                      weight=1, label=None,
+                      percent_variation=.05, streaming_window=1, accuracy=.99,
+                      autotunePV=True, autotuneRange=True, autotune_by_feature=True, autotune_max_clusters=1000, exclusions=None,
+                      streaming_autotune=True, streaming_buffer=10000, learning_numerator=10, learning_denominator=10000, learning_max_clusters=1000, learning_samples=1000000,
+                      config=None):
+
+        """Returns the posted clustering configuration
 
          Args:
              feature_count (int): number of features per vector
              numeric_format (str): numeric type of data (one of "float32", "uint16", or "int16")
              min: list of minimum values per feature, if specified as a single value, use that on all features
              max: list of maximum values per feature, if specified as a single value, use that on all features
-             weight (float):
-             labels (list):
-             percent_variation (float):
-             streaming_window (integer):
-             accuracy (float):
-             config (dict):
+             weight (float): influence each column has on creating a new cluster
+             labels (list): name of each feature (if applicable)
+             percent_variation (float): amount of variation within each cluster
+             streaming_window (integer): number of consecutive vectors treated as one inference (parametric parameter)
+             accuracy (float): statistical accuracy of the clusters
+             autotunePV (bool): whether to autotune the percent variation
+             autotuneRange (bool): whether to autotune the min and max values
+             autotune_by_feature (bool): whether to have individually set min and max values for each feature
+             autotune_max_clusters (int): max number of clusters allowed
+             exclusions (list): features to exclude while autotuning
+             streaming_autotune (bool): whether to autotune while in streaming mode
+             streaming_buffer (int): number of samples to autotune on
+             learning_numerator (int): max number of new clusters learned
+             learning_denominator (int): number of samples over which the new clusters are learned
+             learning_max_clusters (int): max number of clusters before turning off learning
+             learning_samples (int): max number of samples before turning off learning
+             config (dict): dictionary of configuration parameters
 
          Returns:
              result (boolean): true if successful (configuration was successfully loaded into nano pod instance)
@@ -229,6 +331,12 @@ class NanoHandle:
 
          """
 
+        if config is None:
+            config = self.create_config(feature_count, numeric_format, cluster_mode, min_val, max_val, weight, label,
+                                        percent_variation, streaming_window, accuracy,
+                                        autotunePV, autotuneRange, autotune_by_feature, autotune_max_clusters, exclusions,
+                                        streaming_autotune, streaming_buffer, learning_numerator, learning_denominator, learning_max_clusters, learning_samples)
+
         body = json.dumps(config)
 
         config_cmd = self.url + 'clusterConfig/' + self.instance + '?api-tenant=' + self.api_tenant
@@ -239,7 +347,7 @@ class NanoHandle:
         return result, response
 
     def nano_list(self):
-        """returns list of nano instances allocated for a pod
+        """Returns list of nano instances allocated for a pod
 
         Returns:
             result (boolean):  true if successful (list was returned)
@@ -252,6 +360,7 @@ class NanoHandle:
 
         return simple_get(self, instance_cmd)
 
+    @_is_configured
     def save_nano(self, filename):
         """serialize a nano pod instance and save to a local file
 
@@ -282,7 +391,7 @@ class NanoHandle:
         return True, None
 
     def restore_nano(self, filename):
-        """restore a nano pod instance from local file
+        """Restore a nano pod instance from local file
 
         Args:
             filename (str): path to local file containing saved pod instance
@@ -296,9 +405,9 @@ class NanoHandle:
         # verify that input file is a valid nano file (gzip'd tar with Magic Number)
         try:
             with tarfile.open(filename, 'r:gz') as tp:
-                with tp.extractfile('BoonNano/MagicNumber') as magic_fp:
+                with tp.extractfile('/CommonState/MagicNumber') as magic_fp:
                     magic_num = magic_fp.read()
-                    if magic_num != b'\xef\xbe':
+                    if magic_num != b'\xda\xba':
                         return False, 'file {} is not a Boon Logic nano-formatted file, bad magic number'.format(
                             filename)
         except KeyError:
@@ -323,14 +432,9 @@ class NanoHandle:
 
         return True, response
 
-    def autotune_config(self, autotune_pv=True, autotune_range=True, by_feature=False, exclusions=None):
-        """autotunes the percent variation, min and max for each feature
-
-        Args:
-            autotune_pv (boolean):
-            autotune_range (boolean):
-            by_feature (boolean):
-            exclusions (list):
+    @_is_configured
+    def autotune_config(self):
+        """Autotunes the percent variation, min and max for each feature
 
         Returns:
             result (boolean): true if successful (autotuning was completed)
@@ -339,20 +443,14 @@ class NanoHandle:
         """
 
         # build command
-        config_cmd = self.url + 'autoTuneConfig/' + self.instance + '?api-tenant=' + self.api_tenant
-        config_cmd += '&byFeature=' + str(by_feature).lower()
-        config_cmd += '&autoTunePV=' + str(autotune_pv).lower()
-        config_cmd += '&autoTuneRange=' + str(autotune_range).lower()
-        if isinstance(exclusions, list):
-            config_cmd += '&exclusions=' + ",".join([str(s) for s in exclusions])
-        elif exclusions:
-            return False, 'exclusions must be a list'
+        config_cmd = self.url + 'autoTune/' + self.instance + '?api-tenant=' + self.api_tenant
 
         # autotune parameters
         return simple_post(self, config_cmd)
 
+    @_is_configured
     def get_config(self):
-        """gets the configuration for this nano pod instance
+        """Gets the configuration for this nano pod instance
 
         Returns:
             result (boolean): true if successful (configuration was found)
@@ -362,8 +460,9 @@ class NanoHandle:
         config_cmd = self.url + 'clusterConfig/' + self.instance + '?api-tenant=' + self.api_tenant
         return simple_get(self, config_cmd)
 
+    @_is_configured
     def load_file(self, file, file_type, gzip=False, append_data=False):
-        """load nano data from a file
+        """Load nano data from a file
 
         Args:
             file (str): local path to data file
@@ -403,8 +502,9 @@ class NanoHandle:
 
         return multipart_post(self, dataset_cmd, fields=fields)
 
+    @_is_configured
     def load_data(self, data, append_data=False):
-        """load nano data from an existing numpy array or simple python list
+        """Load nano data from an existing numpy array or simple python list
 
         Args:
             data (np.ndarray or list): numpy array or list of data values
@@ -416,22 +516,7 @@ class NanoHandle:
             response (str): None when result is true, error string when result=false
 
         """
-
-        if not isinstance(data, np.ndarray):
-            if self.numeric_format == 'int16':
-                data = np.asarray(data, dtype=np.int16)
-            elif self.numeric_format == 'float32':
-                data = np.asarray(data, dtype=np.float32)
-            elif self.numeric_format == 'uint16':
-                data = np.asarray(data, dtype=np.uint16)
-
-        if self.numeric_format == 'int16':
-            data = data.astype(np.int16)
-        elif self.numeric_format == 'float32':
-            data = data.astype(np.float32)
-        elif self.numeric_format == 'uint16':
-            data = data.astype(np.uint16)
-        data = data.tostring()
+        data = normalize_nano_data(data, self.numeric_format)
         file_name = 'dummy_filename.bin'
         file_type = 'raw'
 
@@ -444,8 +529,25 @@ class NanoHandle:
 
         return multipart_post(self, dataset_cmd, fields=fields)
 
+    def set_learning_status(self, status):
+        """returns list of nano instances allocated for a pod
+
+        Args:
+            status (boolean): true or false of whether to learning is on or off
+
+        Returns:
+            result (boolean):  true if successful (list was returned)
+            response (str): json dictionary of pod instances when result=true, error string when result=false
+
+        """
+
+        # build command
+        learning_cmd = self.url + 'learning/' + self.instance + '?enable=' + str(status).lower() + '&api-tenant=' + self.api_tenant
+
+        return simple_post(self, learning_cmd)
+
     def run_nano(self, results=None):
-        """ clusters the data in the nano pod buffer and returns the specified results
+        """Clusters the data in the nano pod buffer and returns the specified results
 
         Args:
             results (str): comma separated list of result specifiers
@@ -460,6 +562,8 @@ class NanoHandle:
 
                 DI = distance index
 
+                All = ID,SI,RI,FI,DI
+
         Returns:
             result (boolean): true if successful (nano was successfully run)
             response (dict or str): dictionary of results when result is true, error message when result = false
@@ -482,8 +586,56 @@ class NanoHandle:
 
         return simple_post(self, nano_cmd)
 
+    @_is_configured
+    def run_streaming_nano(self, data, results=None):
+        """Load streaming data into self-autotuning nano pod instance, run the nano and return results
+
+        Args:
+            data (np.ndarray or list): numpy array or list of data values
+            results (str): comma separated list of result specifiers
+
+                ID = cluster ID
+
+                SI = smoothed anomaly index
+
+                RI = raw anomaly index
+
+                FI = frequency index
+
+                DI = distance index
+
+                All = ID,SI,RI,FI,DI
+
+        Returns:
+            result (boolean): true if successful (data was successful streamed to nano pod instance)
+            response (dict or str): dictionary of results when result is true, error message when result = false
+
+        """
+        data = normalize_nano_data(data, self.numeric_format)
+        file_name = 'dummy_filename.bin'
+        file_type = 'raw'
+
+        fields = {'data': (file_name, data)}
+
+        results_str = ''
+        if str(results) == 'All':
+            results_str = 'ID,SI,RI,FI,DI'
+        elif results:
+            for result in results.split(','):
+                if result not in ['ID', 'SI', 'RI', 'FI', 'DI']:
+                    return False, 'unknown result "{}" found in results parameter'.format(result)
+            results_str = results
+
+        # build command
+        streaming_cmd = self.url + 'nanoRunStreaming/' + self.instance + '?api-tenant=' + self.api_tenant
+        streaming_cmd += '&fileType=' + file_type
+        if results:
+            streaming_cmd += '&results=' + results_str
+
+        return multipart_post(self, streaming_cmd, fields=fields)
+
     def get_version(self):
-        """ results related to the bytes processed/in the buffer
+        """Version information for this nano pod
 
         Returns:
             result (boolean): true if successful (version information was retrieved)
@@ -495,8 +647,9 @@ class NanoHandle:
         version_cmd = self.url[:-3] + 'version' + '?api-tenant=' + self.api_tenant
         return simple_get(self, version_cmd)
 
+    @_is_configured
     def get_buffer_status(self):
-        """ results related to the bytes processed/in the buffer
+        """Results related to the bytes processed/in the buffer
 
         Returns:
             result (boolean): true if successful (nano was successfully run)
@@ -506,23 +659,24 @@ class NanoHandle:
         status_cmd = self.url + 'bufferStatus/' + self.instance + '?api-tenant=' + self.api_tenant
         return simple_get(self, status_cmd)
 
+    @_is_configured
     def get_nano_results(self, results='All'):
-        """ results per pattern
+        """Results per pattern
 
         Args:
             results (str): comma separated list of results
 
-                ID: cluster ID
+                ID = cluster ID
 
-                SI: smoothed anomaly index
+                SI = smoothed anomaly index
 
-                RI: raw anomaly index
+                RI = raw anomaly index
 
-                FI: frequency index
+                FI = frequency index
 
-                DI: distance index
+                DI = distance index
 
-                All: ID,SI,RI,FI,DI
+                All = ID,SI,RI,FI,DI
 
         """
         # build results command
@@ -540,8 +694,9 @@ class NanoHandle:
 
         return simple_get(self, results_cmd)
 
+    @_is_configured
     def get_nano_status(self, results='All'):
-        """results in relation to each cluster/overall stats
+        """Results in relation to each cluster/overall stats
 
         Args:
             results (str): comma separated list of results
@@ -558,8 +713,6 @@ class NanoHandle:
 
                 distanceIndexes = distance index (includes 0 cluster)
 
-                patternMemory = base64 pattern memory (overall)
-
                 totalInferences = total number of patterns clustered (overall)
 
                 averageInferenceTime = time in milliseconds to cluster per
@@ -567,6 +720,8 @@ class NanoHandle:
 
                 numClusters = total number of clusters (includes 0 cluster) (overall)
 
+                All = PCA,clusterGrowth,clusterSizes,anomalyIndexes,frequencyIndexes,distanceIndexes,totalInferences,numClusters
+
         Returns:
             result (boolean): true if successful (nano was successfully run)
             response (dict or str): dictionary of results when result is true, error message when result = false
@@ -589,3 +744,20 @@ class NanoHandle:
         results_cmd = results_cmd + '&results=' + results_str
 
         return simple_get(self, results_cmd)
+
+def normalize_nano_data(data, numeric_format):
+    # Whatever type data comes in as, cast it to numpy array
+    data = np.asarray(data)
+
+    # Cast numpy array to correct numeric type for serialization
+    if numeric_format == 'int16':
+        data = data.astype(np.int16)
+    elif numeric_format == 'float32':
+        data = data.astype(np.float32)
+    elif numeric_format == 'uint16':
+        data = data.astype(np.uint16)
+
+    # Serialize to binary blob
+    data = data.tostring()
+
+    return data
